from dataclasses import dataclass
from .enums import PriceType

class Tier:
    lower_peak_limit: float
    value: float


class TieredPrice:
    values: list[Tier]


@dataclass
class LocalePrice:
    price_type: PriceType
    value: float | TieredPrice
    currency:str




#if tiered, do a list of the tiers
    


from ..querytypes.const import (
    QUERYTYPE_AVERAGEOFTHREEDAYS,
    QUERYTYPE_AVERAGEOFTHREEDAYS_MIN,
    QUERYTYPE_AVERAGEOFTHREEHOURS_MIN, QUERYTYPE_AVERAGEOFTHREEHOURS,
    QUERYTYPE_BASICMAX
)
from ..querytypes.querytypes import QUERYTYPES
from ..locale_model import Locale_Type
from dataclasses import dataclass

@dataclass(frozen=True)
class NO_Tensio(Locale_Type):
    observed_peak = QUERYTYPE_AVERAGEOFTHREEDAYS_MIN
    charged_peak = QUERYTYPE_AVERAGEOFTHREEDAYS
    query_model = QUERYTYPES[QUERYTYPE_AVERAGEOFTHREEDAYS]

@dataclass(frozen=True)
class NO_LNett(Locale_Type):
    observed_peak = QUERYTYPE_AVERAGEOFTHREEHOURS_MIN
    charged_peak = QUERYTYPE_AVERAGEOFTHREEHOURS
    query_model = QUERYTYPES[QUERYTYPE_AVERAGEOFTHREEHOURS]

#docs: https://www.l-nett.no/nynettleie/slik-blir-ny-nettleie-og-pris        


@dataclass(frozen=True)
class NO_GlitreEnergi(Locale_Type):
    observed_peak = QUERYTYPE_BASICMAX
    charged_peak = QUERYTYPE_BASICMAX
    query_model = QUERYTYPES[QUERYTYPE_BASICMAX]

#docs: https://www.glitreenergi-nett.no/smart-nettleie/


@dataclass(frozen=True)
class NO_AgderEnergi(Locale_Type):
    observed_peak = QUERYTYPE_AVERAGEOFTHREEDAYS_MIN
    charged_peak = QUERYTYPE_AVERAGEOFTHREEDAYS
    query_model = QUERYTYPES[QUERYTYPE_AVERAGEOFTHREEDAYS]


@dataclass(frozen=True)
class NO_Elvia(Locale_Type):
    observed_peak = QUERYTYPE_AVERAGEOFTHREEDAYS_MIN
    charged_peak = QUERYTYPE_AVERAGEOFTHREEDAYS
    query_model = QUERYTYPES[QUERYTYPE_AVERAGEOFTHREEDAYS]


@dataclass(frozen=True)
class NO_Lede(Locale_Type):
    observed_peak = QUERYTYPE_AVERAGEOFTHREEDAYS_MIN
    charged_peak = QUERYTYPE_AVERAGEOFTHREEDAYS
    query_model = QUERYTYPES[QUERYTYPE_AVERAGEOFTHREEDAYS]


@dataclass(frozen=True)
class NO_BKK(Locale_Type):
    observed_peak = QUERYTYPE_AVERAGEOFTHREEDAYS_MIN
    charged_peak = QUERYTYPE_AVERAGEOFTHREEDAYS
    query_model = QUERYTYPES[QUERYTYPE_AVERAGEOFTHREEDAYS]

@dataclass(frozen=True)
class NO_Mellom(Locale_Type):
    observed_peak = QUERYTYPE_AVERAGEOFTHREEDAYS_MIN
    charged_peak = QUERYTYPE_AVERAGEOFTHREEDAYS
    query_model = QUERYTYPES[QUERYTYPE_AVERAGEOFTHREEDAYS]


@dataclass(frozen=True)
class NO_AskerNett(Locale_Type):
    observed_peak = QUERYTYPE_AVERAGEOFTHREEDAYS_MIN
    charged_peak = QUERYTYPE_AVERAGEOFTHREEDAYS
    query_model = QUERYTYPES[QUERYTYPE_AVERAGEOFTHREEDAYS]
    
#docs: https://askernett.no/ny-nettleiemodell-for-alle-fra-1-juli-2022/



from enum import Enum
from const import (
    CAUTIONHOURTYPE_SUAVE,
    CAUTIONHOURTYPE_INTERMEDIATE,
    CAUTIONHOURTYPE_AGGRESSIVE
)

class CautionHourType(Enum):
    Suave = CAUTIONHOURTYPE_SUAVE
    Intermediate = CAUTIONHOURTYPE_INTERMEDIATE
    Aggressive = CAUTIONHOURTYPE_AGGRESSIVE


ttt = CautionHourType(CAUTIONHOURTYPE_SUAVE)
#detta är enumvärdet
print(ttt.name)
#detta är strängen
print(ttt.value)


from dataclasses import dataclass, field
from typing import List
from .enums import Dividents

@dataclass
class datepart_model:
    type: str = field(default_factory=lambda : "")
    dttype: str = field(default_factory=lambda : "")
    values: List[int] = field(default_factory=lambda : [])

@dataclass
class group:
    divident: Dividents = field(default_factory=lambda: Dividents.UNSET)
    dateparts: List[datepart_model] = field(default_factory=lambda : [datepart_model()])

@dataclass
class queryservicemodel:
    groups: List[group] = field(default_factory=lambda : [group()])



from dataclasses import dataclass, field
from typing import List, Dict
from .hourobject import HourObject

@dataclass(frozen=False)
class HoursModel:
    non_hours: List[int] = field(default_factory=lambda : [])
    caution_hours: List[int] = field(default_factory=lambda : [])
    dynamic_caution_hours: Dict[int, float] = field(default_factory=lambda : {})
    hours_today: HourObject = field(default_factory=lambda : HourObject([],[],dict()))
    hours_tomorrow: HourObject = field(default_factory=lambda : HourObject([],[],dict()))

    def update_non_hours(
        self, 
        hour:int
        ) -> None:
        ret = []
        ret.extend(h for h in self.hours_today.nh if h >= hour)
        ret.extend(h for h in self.hours_tomorrow.nh if h < hour)
        self.non_hours = ret
    
    def update_caution_hours(
        self, 
        hour:int
        ) -> None:
        ret = []
        ret.extend(h for h in self.hours_today.ch if h >= hour)
        ret.extend(h for h in self.hours_tomorrow.ch if h < hour)
        self.caution_hours = ret

    def update_dynanmic_caution_hours(
        self, 
        hour:int
        ) -> None:
        ret = {}
        ret.update({k: v for k, v in self.hours_today.dyn_ch.items() if k >= hour})
        ret.update({k: v for k, v in self.hours_tomorrow.dyn_ch.items() if k < hour})
        self.dynamic_caution_hours = ret


@dataclass(frozen=False)
class HourSelectionOptions:
    cautionhour_type: float = 0
    absolute_top_price: float = 0
    min_price: float = 0

    @staticmethod
    def set_absolute_top_price(val) -> float:
        if val is None:
            return float("inf")
        if val <= 0:
            return float("inf")
        return float(val)


@dataclass(frozen=False)
class HourSelectionModel:
    prices_today: List[float] = field(default_factory=lambda : [])
    prices_tomorrow: List[float] = field(default_factory=lambda : [])
    hours: HoursModel = HoursModel()
    options: HourSelectionOptions = HourSelectionOptions

    def validate(self):
        assert 0 < self.options.cautionhour_type <= 1


from dataclasses import dataclass

@dataclass
class HourObject:
    nh: list
    ch: list
    dyn_ch: dict


@dataclass
class HourObjectExtended(HourObject):
    pricedict: dict

    def remove_cheap_hours(self, min_price:float = 0) -> HourObject:
            lst = (h for h in self.pricedict if self.pricedict[h] < min_price)
            for h in lst:
                if h in self.nh:
                    self.nh.remove(h)
                elif h in self.ch:
                    self.ch.remove(h)
                    self.dyn_ch.pop(h)    
            return HourObject(self.nh, self.ch, self.dyn_ch)

    def add_expensive_hours(self, max_price: float = 0) -> HourObject:
        lst = (h for h in self.pricedict if self.pricedict[h] >= max_price)
        for h in lst:
            if h not in self.nh:
                self.nh.append(h)
                if h in self.ch:
                    self.ch.remove(h)
                if len(self.dyn_ch) > 0:
                    if h in self.dyn_ch.keys():
                        self.dyn_ch.pop(h)
        self.nh.sort()
        return HourObject(self.nh, self.ch, self.dyn_ch)


from dataclasses import dataclass
from datetime import datetime
import logging

_LOGGER = logging.getLogger(__name__)

@dataclass
class Day:
    sessions: int
    charge: float


class EnergyWeekly:
    model: dict[int, Day]

    def __init__(self, incoming_dict: dict = None):
        if incoming_dict is not None:
            self.unpack(incoming_dict)
        else:
            self.set_init_model()

    def set_init_model(self):
        ret = {}
        for i in range(0,6):
            ret[i] = Day(0,0)
        self.model = ret

    @property
    def export(self) -> dict[int, dict[str, float]]:
        "export the model to HA-friendly states"
        ret = {}
        for idx, m in self.model.items():
            m_ret = {
                "sessions": m.sessions,
                "total_charge": m.charge
            }
            ret[idx] = m_ret

    @property
    def average(self) -> float:
        sessions = sum([h.sessions for h in self.model.values()])
        charge = sum([h.charge for h in self.model.values()])
        try:
            return round(charge/sessions,1)
        except:
            return 0.0
        
    @property
    def total_sessions(self) -> int:
        try:
            return sum([h.sessions for h in self.model.values()])
        except:
            _LOGGER.debug("could not retreive sessions.")
            return 0
        
    @property
    def total_charge(self) -> float:
        try:
            return sum([h.charge for h in self.model.values()])
        except:
            _LOGGER.debug("could not retreive total charge")
            return 0.0

    def unpack(self, incoming: dict) -> dict[int, Day]:
        if isinstance(incoming, dict) and len(incoming) > 0:
            model = {}
            for idx, m in incoming.items():
                m_ret = Day(m["sessions"], m["total_charge"])
                model[idx] = m_ret
            self.model = model

    def update(self, charge:float):
        if charge > 0:
            self.model[datetime.now().weekday()].charge += charge
            self.model[datetime.now().weekday()].sessions += 1

    def average_for_day(self, day: int) -> float:
        try:
            return self.model[day].charge/self.model[day].sessions
        except:
            return 0.0

# EXAMPLE = {
#     0: Day(1,50),
#     1: Day(4,50),
#     2: Day(2,15),
#     3: Day(3,20),
#     4: Day(10,123),
#     5: Day(0,0),
#     6: Day(10, 110)
# }

# EXPORT_EXAMPLE = {0: {'sessions': 1, 'total_charge': 50}, 1: {'sessions': 4, 'total_charge': 50}, 2: {'sessions': 2, 'total_charge': 15}, 3: {'sessions': 3, 'total_charge': 20}, 4: {'sessions': 10, 'total_charge': 123}, 5: {'sessions': 0, 'total_charge': 0}, 6: {'sessions': 10, 'total_charge': 110}}

# e = EnergyWeekly()
# print(e.model)
# print(e.average)
# print(e.average_for_day(0))


from ..services.session.session import SessionPrice
import pytest

def test_session_fluctuate():
    s = SessionPrice()
    timer = 1651607299
    s._set_delta(timer)
    s.update_price(0.6, timer)
    s.update_power_reading(6000, timer)    
    timer += 1200
    s.update_price(0.3, timer)
    s.update_power_reading(3000, timer)    
    timer += 1200    
    s.terminate(timer)
    assert s.total_energy == 3
    assert round(s.total_price,4) == 1.5

def test_session_fluctuate_tenfold():
    s = SessionPrice()
    timer = 1651607299
    s._set_delta(timer)
    s.update_price(6, timer)
    s.update_power_reading(6000, timer)    
    timer += 1200
    s.update_price(3, timer)
    s.update_power_reading(3000, timer)    
    timer += 1200    
    s.terminate(timer)
    assert s.total_energy == 3
    assert round(s.total_price,4) == 15

def test_session_split_price():
    s = SessionPrice()
    timer = 1651607299
    s._set_delta(timer)
    s.update_price(2, timer)
    s.update_power_reading(2000, timer)
    timer += 900
    s.update_price(3, timer)
    timer += 900
    s.update_power_reading(1000, timer)
    timer += 1800
    s.terminate(timer)

    assert s.total_energy == 1.5
    assert s.total_price == 4


def test_session_full_hour():
    s = SessionPrice()
    timer = 1651607299
    s._set_delta(timer)
    s.update_price(1, timer)
    s.update_power_reading(1000, timer)
    timer += 3600
    s.update_power_reading(1000, timer)
    s.terminate(timer)
    assert s.total_energy == 1
    assert s.total_price == 1

def test_session_full_hour_no_price():
    s = SessionPrice()
    timer = 1651607299
    s._set_delta(timer)
    s.update_power_reading(1000, timer)
    timer += 3600
    s.update_power_reading(1000, timer)
    #s.terminate(timer)
    assert s.total_energy == 1

def test_session_half_hour():
    s = SessionPrice()
    timer = 1651607299
    s._set_delta(timer)
    s.update_price(1, timer)
    s.update_power_reading(1000, timer)
    timer += 1800
    s.update_power_reading(1000, timer)
    s.terminate(timer)
    assert s.total_energy == 0.5
    assert s.total_price == 0.5

def test_session_with_zero_periods():
    s = SessionPrice()
    timer = 1651607299
    s._set_delta(timer)
    s.update_price(2, timer)
    s.update_power_reading(4000, timer)
    timer += 900
     #2kr
    s.update_power_reading(0, timer) 
    timer += 900
     #2kr
    s.update_price(3, timer)
    timer += 900
     #2kr
    s.update_power_reading(1000, timer)
    timer += 1800
     #3.5kr
    s.terminate(timer)

    assert s.total_energy == 1.5
    assert s.total_price == 3.5

def test_session_with_zero_periods_price_update_in_between():
    s = SessionPrice()
    timer = 1651607299
    s._set_delta(timer)
    s.update_price(2, timer)
    s.update_power_reading(4000, timer)
    timer += 900
     #2kr
    s.update_price(3, timer)
    timer += 900
    #5kr
    s.update_power_reading(0, timer) 
    timer += 900
     #5kr
    s.update_power_reading(1000, timer)
    timer += 1800
     #6.5kr
    s.terminate(timer)

    assert s.total_energy == 2.5
    assert s.total_price == 6.5

def test_session_get_status():
    s = SessionPrice()
    timer = 1651607299
    s._set_delta(timer)
    s.update_price(2, timer)
    s.update_power_reading(4000, timer)
    timer += 900
    status = s.get_status()
    assert status["price"] == 0
    s.update_price(3, timer)
    timer += 900
    status = s.get_status()
    assert status["price"] == 2
    s.update_power_reading(0, timer) 
    timer += 900
    status = s.get_status()
    assert status["price"] == 5
     #5kr
    s.update_power_reading(1000, timer)
    timer += 1800
    status = s.get_status()
    assert status["price"] == 5
    assert status["energy"]["value"] == 2
    s.terminate(timer)

    assert s.total_energy == 2.5
    assert s.total_price == 6.5






import logging
from ...util import nametoid

_LOGGER = logging.getLogger(__name__)


class HubMember:
    def __init__(self, data_type, listenerentity = None, initval = None, name = None, init_override:bool = False):
        self._value = initval
        self._type = data_type
        self._listenerentity = listenerentity
        self.name = name
        self.id = nametoid(self.name) if self.name is not None else None
        self.warned_not_initialized = False
        self._is_initialized = init_override

    @property
    def is_initialized(self) -> bool:
        if self._is_initialized:
            return True
        if isinstance(self.value, self._type):
            _LOGGER.debug(f"{self._listenerentity} has initialized")
            self._is_initialized = True
            return True
        if not self.warned_not_initialized:
            _LOGGER.error(f"{self._listenerentity} was not {self._type}. {self.value}")
            self.warned_not_initialized = True
        return False

    @property
    def entity(self) -> str:
        return self._listenerentity

    @entity.setter
    def entity(self, val: str):
        self._listenerentity = val

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        if isinstance(value, self._type):
            self._value = value
        elif self._type is float:
            try:
                self._value = float(value)
            except ValueError:
                self._value = 0
        elif self._type is int:
            try:
                self._value = int(float(value))
            except ValueError:
                self._value = 0
        elif self._type is bool:
            try:
                if value is None:
                    self._value = False
                elif value.lower() == "on":
                    self._value = True
                elif value.lower() == "off":
                    self._value = False
            except ValueError as e:
                msg = f"Could not parse bool, setting to false to be sure {value}, {self._listenerentity}, {e}"
                _LOGGER.error(msg)
                self._value = False
        elif  self._type is str:
            self._value = str(value)



from datetime import datetime
#from ...hub.hub import HubBase
from ...models.chargerstates import CHARGERSTATES
from .thresholdbase import ThresholdBase


class Threshold(ThresholdBase):
    def __init__(self, hub):
        self._hub = hub
        super().__init__(hub)

    @property
    def allowedcurrent(self) -> int:
        amps = self._setcurrentdict()
        if self._hub.chargecontroller.status is not CHARGERSTATES.Start.name:
            return min(amps.values())
        return ThresholdBase.allowed_current(
            datetime.now().minute,
            self._hub.sensors.powersensormovingaverage.value if self._hub.sensors.powersensormovingaverage.value is not None else 0,
            self._hub.sensors.charger_enabled.value,
            self._hub.sensors.charger_done.value,
            amps,
            self._hub.sensors.totalhourlyenergy.value,
            self._hub.current_peak_dynamic,
            self._hub.sensors.locale.data.is_quarterly(self._hub.sensors.locale.data)
        )



from dataclasses import dataclass, field
from datetime import datetime

from ..locale.free_charge import FreeChargePattern
from ...models.locale.price import LocalePrice
from .querytypes.const import (
HOURLY,
QUARTER_HOURLY
)
from .querytypes.querytypes import(
LocaleQuery
)

@dataclass(frozen=True)
class Locale_Type:
    observed_peak: str
    charged_peak: str
    query_model: LocaleQuery
    price: LocalePrice = None
    free_charge_pattern: FreeChargePattern = None
    peak_cycle: str = HOURLY
    #converted: bool = False #transition key to remove sql-dependency
    is_quarterly: bool = field(init=False, repr=False)
    

    def free_charge(self, mockdt: datetime=datetime.min) -> bool:
        if self.free_charge_pattern is None:
            return False
        now = datetime.now() if mockdt is datetime.min else mockdt
        return self.free_charge_pattern.free_charge(now)

    def is_quarterly(self) -> bool:
        return self.peak_cycle == QUARTER_HOURLY






from datetime import datetime
import pytest

from ..services.scheduler.scheduler import Scheduler
from ..models.hourselection.hourselectionmodels import HourSelectionOptions

MOCKPRICES = [0.422, 0.341, 0.309, 0.322, 0.331, 0.422, 0.773, 1.169, 1.794, 1.975, 2.119, 1.849, 1.309, 1.24, 1.137, 1.122, 1.14, 1.3, 1.397, 1.971, 1.377, 1.178, 1.05, 0.709]
MOCKPRICES_TOMORROW = [0.513, 0.452, 0.458, 0.413, 0.377, 0.498, 0.693, 2.145, 2.193, 2.191, 2.184, 2.168, 2.153, 2.144, 2.135, 2.12, 2.134, 2.156, 2.165, 2.159, 2.146, 2.121, 2.081, 0.983]

def test_scheduler_overnight_1():
    s = Scheduler()
    MOCKDT = datetime.strptime('22-07-20 19:00', '%y-%m-%d %H:%M')
    dep_time_obj = datetime.strptime('22-07-21 06:00', '%y-%m-%d %H:%M')
    s.create(desired_charge=11, departuretime=dep_time_obj, starttime=MOCKDT)
    s._update(avg24=710, peak=4.71, charged_amount=0, prices=MOCKPRICES, prices_tomorrow=MOCKPRICES_TOMORROW,mockdt=MOCKDT)
    assert s.model.non_hours == [19,20,21,22,23,0,2,5]
    assert s.model.hours_charge == {datetime(2022, 7, 21, 1, 0): 0.7, datetime(2022, 7, 21, 3, 0): 1, datetime(2022, 7, 21, 4, 0): 1}

def test_scheduler_overnight_1_maxprice():
    opt = HourSelectionOptions(absolute_top_price=0.4)
    s = Scheduler(opt)
    MOCKDT = datetime.strptime('22-07-20 19:00', '%y-%m-%d %H:%M')
    dep_time_obj = datetime.strptime('22-07-21 06:00', '%y-%m-%d %H:%M')
    s.create(desired_charge=11, departuretime=dep_time_obj, starttime=MOCKDT)
    s._update(avg24=710, peak=4.71, charged_amount=0, prices=MOCKPRICES, prices_tomorrow=MOCKPRICES_TOMORROW,mockdt=MOCKDT)
    assert s.model.non_hours == [19,20,21,22,23,0,1,2,3,5]
    assert s.model.hours_charge == {datetime(2022, 7, 21, 4, 0): 1}

def test_scheduler_cancel_on_departure():
    s = Scheduler()
    MOCKDT = datetime.strptime('22-07-20 19:00', '%y-%m-%d %H:%M')
    dep_time_obj = datetime.strptime('22-07-21 06:00', '%y-%m-%d %H:%M')
    s.create(desired_charge=11, departuretime=dep_time_obj, starttime=MOCKDT)
    s._update(avg24=710, peak=4.71, charged_amount=0, prices=MOCKPRICES, prices_tomorrow=MOCKPRICES_TOMORROW,mockdt=MOCKDT)
    s._update(avg24=310, peak=4.71, charged_amount=0, prices=MOCKPRICES, prices_tomorrow=MOCKPRICES_TOMORROW,mockdt=dep_time_obj)
    assert s.scheduler_active is False

def test_scheduler_overnight_2():
    s = Scheduler()
    MOCKDT = datetime.strptime('22-07-20 19:00', '%y-%m-%d %H:%M')
    midnight_dt = datetime.strptime('22-07-21 01:00', '%y-%m-%d %H:%M')
    dep_time_obj = datetime.strptime('22-07-21 06:00', '%y-%m-%d %H:%M')
    s.create(desired_charge=11, departuretime=dep_time_obj, starttime=MOCKDT)
    s._update(avg24=710, peak=4.71, charged_amount=0, prices=MOCKPRICES, prices_tomorrow=MOCKPRICES_TOMORROW,mockdt=MOCKDT)
    assert s.model.non_hours == [19,20,21,22,23,0,2,5]
    assert s.model.hours_charge == {datetime(2022, 7, 21, 1, 0): 0.7, datetime(2022, 7, 21, 3, 0): 1, datetime(2022, 7, 21, 4, 0): 1}
    s._update(avg24=710, peak=4.71, charged_amount=0, prices=MOCKPRICES_TOMORROW,mockdt=midnight_dt)
    assert s.model.non_hours == [2,5]

def test_scheduler_overnight_3():
    s = Scheduler()
    start_dt = datetime.strptime('22-07-20 11:00', '%y-%m-%d %H:%M')
    dep_time_dt = datetime.strptime('22-07-21 09:00', '%y-%m-%d %H:%M')
    s.create(desired_charge=9, departuretime=dep_time_dt, starttime=start_dt)
    s._update(avg24=410, peak=1.5, charged_amount=0, prices=MOCKPRICES,mockdt=start_dt)
    assert s.model.non_hours == [11,18,19,20]
    s._update(avg24=410, peak=1.5, charged_amount=1, prices=MOCKPRICES,mockdt=datetime.strptime('22-07-20 12:05', '%y-%m-%d %H:%M'))
    s._update(avg24=610, peak=1.5, charged_amount=0.7, prices=MOCKPRICES, prices_tomorrow=MOCKPRICES_TOMORROW,mockdt=datetime.strptime('22-07-20 13:01', '%y-%m-%d %H:%M'))
    assert s.model.non_hours == [14, 15, 16, 17, 18,19,20,21,7,8]
    s._update(avg24=610, peak=1.5, charged_amount=0.7, prices=MOCKPRICES, prices_tomorrow=MOCKPRICES_TOMORROW,mockdt=datetime.strptime('22-07-20 22:11', '%y-%m-%d %H:%M'))
    assert s.model.non_hours == [7,8]
    s._update(avg24=310, peak=1.5, charged_amount=1, prices=MOCKPRICES_TOMORROW,mockdt=datetime.strptime('22-07-21 00:03', '%y-%m-%d %H:%M'))
    assert s.model.non_hours == [6,7,8]

def test_scheduler_overnight_no_charge():
    """this is fine. since there is no charge happening, the non-hours will be less and less as time to departure gets smaller"""
    s = Scheduler()
    start_dt = datetime.strptime('22-07-20 11:00', '%y-%m-%d %H:%M')
    dep_time_dt = datetime.strptime('22-07-21 09:00', '%y-%m-%d %H:%M')
    s.create(desired_charge=9, departuretime=dep_time_dt, starttime=start_dt)
    s._update(avg24=310, peak=1.5, charged_amount=0, prices=MOCKPRICES,mockdt=start_dt)
    assert s.model.non_hours == [11,12,18,19,20]
    s._update(avg24=310, peak=1.5, charged_amount=0, prices=MOCKPRICES,mockdt=datetime.strptime('22-07-20 12:05', '%y-%m-%d %H:%M'))
    s._update(avg24=310, peak=1.5, charged_amount=0, prices=MOCKPRICES, prices_tomorrow=MOCKPRICES_TOMORROW,mockdt=datetime.strptime('22-07-20 13:01', '%y-%m-%d %H:%M'))
    assert s.model.non_hours == [14,15,16, 17, 18,19,20,21,22,7,8]
    s._update(avg24=310, peak=1.5, charged_amount=0, prices=MOCKPRICES, prices_tomorrow=MOCKPRICES_TOMORROW,mockdt=datetime.strptime('22-07-20 22:11', '%y-%m-%d %H:%M'))
    assert s.model.non_hours == [7,8]
    s._update(avg24=310, peak=1.5, charged_amount=0, prices=MOCKPRICES_TOMORROW,mockdt=datetime.strptime('22-07-21 00:03', '%y-%m-%d %H:%M'))
    assert s.model.non_hours == [8]


import logging
from datetime import datetime
import statistics as stat
from ....models.hourselection.const import (
    CAUTIONHOURTYPE_SUAVE,
    CAUTIONHOURTYPE_INTERMEDIATE,
    CAUTIONHOURTYPE
)
from .hoursselection_helpers import HourSelectionHelpers as helpers
from .hoursselection_helpers import HourSelectionInterimUpdate as interim
from .hoursselection_helpers import HourSelectionCalculations as calc
from ....models.hourselection.hourobject import HourObject, HourObjectExtended
from ....models.hourselection.hourselectionmodels import HourSelectionModel
from ....models.hourselection.hourtypelist import HourTypeList

_LOGGER = logging.getLogger(__name__)


class HourSelectionService:
    def __init__(self,
    model: HourSelectionModel, base_mock_hour: int = None):
        self.model = model
        self._base_mock_hour = base_mock_hour

    def update(
        self, 
        testhour:int = None
        ):
        if testhour is not None:
            self._base_mock_hour = testhour
        hours_ready = self._update_per_day(self.model.prices_today)
        hours = self._add_remove_limited_hours(hours_ready)
        hours_tomorrow = HourObject([],[],dict())
        if self.model.prices_tomorrow is not None and len(self.model.prices_tomorrow) > 0:
            self.model.hours.conserve_top_up = False
            hours_tomorrow = self._add_remove_limited_hours(
                self._update_per_day(self.model.prices_tomorrow)
                )
            hours, hours_tomorrow, conserve = interim.interim_avg_update(
                today=hours, 
                tomorrow=hours_tomorrow, 
                model =self.model
                )
        self.model.hours.hours_today = hours
        self.model.hours.hours_tomorrow = hours_tomorrow
        self.update_hour_lists(testhour=testhour)
 
    def _update_per_day(self, prices: list) -> HourObjectExtended:
        pricedict = dict
        if prices is not None and len(prices) > 1:
            pricedict = helpers._create_dict(prices)
            normalized_pricedict = helpers._create_dict(
                calc.normalize_prices(prices)
                )
            if stat.stdev(prices) > 0.05:
                prices_ranked = calc.rank_prices(pricedict, normalized_pricedict)
                ready_hours = self._determine_hours(prices_ranked, prices)
                return HourObjectExtended(
                    ready_hours.nh, 
                    ready_hours.ch, 
                    ready_hours.dyn_ch, 
                    pricedict
                    )
            return HourObjectExtended([], [], dict(), pricedict)

    def update_hour_lists(
        self, 
        testhour:int = None, 
        listtype:HourTypeList = None,
        ) -> None:
        hour = self.set_hour(testhour)
        if listtype is not None:
            match listtype:
                case HourTypeList.NonHour:
                    self.model.hours.update_non_hours(hour)
                case HourTypeList.CautionHour:
                    self.model.hours.update_caution_hours(hour)
                case HourTypeList.DynCautionHour:
                    self.model.hours.update_dynanmic_caution_hours(hour)   
                case _:
                    pass
        else:
            self.model.hours.update_non_hours(hour)
            self.model.hours.update_caution_hours(hour)
            self.model.hours.update_dynanmic_caution_hours(hour)

    def _add_remove_limited_hours(self, hours: HourObjectExtended) -> HourObject:
        """Removes cheap hours and adds expensive hours set by user limitation"""
        if hours is None:
            return HourObject([],[],dict())
        if self.model.options.absolute_top_price is not None:
                ret = hours.add_expensive_hours(self.model.options.absolute_top_price)
        else: 
            ret = HourObject(hours.nh, hours.ch, hours.dyn_ch)
        if self.model.options.min_price > 0:
            ret = hours.remove_cheap_hours(self.model.options.min_price)
        return ret

    def _determine_hours(self, price_list: dict, prices: list) -> HourObject:
        _nh = []
        _dyn_ch = {}
        _ch = []
        for p in price_list:
            _permax = self._set_permax(price_list[p]["permax"])
            if float(price_list[p]["permax"]) <= self.model.options.cautionhour_type or float(price_list[p]["val"]) <= (sum(prices)/len(prices)):
                _ch.append(p)
                _dyn_ch[p] = round(_permax,2)
            else:
                _nh.append(p)
        return HourObject(_nh, _ch, _dyn_ch)

    def _set_permax(self, price_input) -> float:
        _permax = round(abs(price_input - 1), 2)
        if self.model.options.cautionhour_type == CAUTIONHOURTYPE[CAUTIONHOURTYPE_SUAVE]:
            _permax += 0.15
        elif self.model.options.cautionhour_type == CAUTIONHOURTYPE[CAUTIONHOURTYPE_INTERMEDIATE]:
            _permax += 0.05
        return _permax

    def set_hour(self, testhour:int = None) -> int:
        return testhour if testhour is not None else self._base_mock_hour if self._base_mock_hour is not None else datetime.now().hour



from ..free_charge import FreeChargePattern
from ....models.locale.enums import CalendarPeriods
from ..querytypes.const import (
    QUERYTYPE_BASICMAX
)

from ..locale_model import Locale_Type
from dataclasses import dataclass

@dataclass(frozen=True)
class Default(Locale_Type):
    observed_peak = QUERYTYPE_BASICMAX
    charged_peak = QUERYTYPE_BASICMAX


@dataclass(frozen=True)
class NoPeak(Locale_Type):
    observed_peak = QUERYTYPE_BASICMAX
    charged_peak = QUERYTYPE_BASICMAX
    free_charge_pattern = FreeChargePattern([
        {
            CalendarPeriods.Month: [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
            CalendarPeriods.Weekday: [0, 1, 2, 3, 4, 5, 6],
            CalendarPeriods.Hour: [0, 1, 2, 3, 4, 5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23]
        }
    ])


import statistics as stat
from typing import Tuple
import logging
from ....models.hourselection.hourobject import HourObject
from ....models.hourselection.hourselectionmodels import HourSelectionModel

_LOGGER = logging.getLogger(__name__)

class HourSelectionInterimUpdate:
    @staticmethod
    def interim_avg_update(today: HourObject, tomorrow: HourObject, model: HourSelectionModel) -> Tuple[HourObject, HourObject, bool]:
        avg = HourSelectionInterimUpdate._get_average_price(model.prices_today, model.prices_tomorrow)
        _today = HourSelectionInterimUpdate._set_interim_per_day(
            True,
            avg, 
            model.prices_today, 
            today, 
            model.options.absolute_top_price, 
            model.options.min_price
            )
        _tomorrow = HourSelectionInterimUpdate._set_interim_per_day(
            False,
            avg, 
            model.prices_tomorrow, 
            tomorrow,  
            model.options.absolute_top_price, 
            model.options.min_price
            )

        conserve = True if _today[1] is True or _tomorrow[1] is True else False
        return _today[0], _tomorrow[0], conserve

    @staticmethod
    def _set_interim_per_day(
        is_today: bool,
        avg: float, 
        prices: list, 
        hour_obj: HourObject, 
        max_price: float = None, 
        min_price: float = None, 
        ) -> Tuple[HourObject, bool]:
        new_nonhours = []
        new_ok_hours = []
        _max_price = float('inf') if max_price is None else max_price
        _min_price = float('-inf') if min_price is None else min_price
        conserve_top_up = False

        for idx, p in enumerate(prices):
            if (idx >= 14 and is_today) or idx < 14:
                if p > avg or p > _max_price:
                    new_nonhours.append(idx)
                elif p <= avg or p <= _min_price:
                    new_ok_hours.append(idx)
        
        for h in new_nonhours:
            if h not in hour_obj.nh:
                hour_obj.nh.append(h)
                if h in hour_obj.ch:
                    hour_obj.ch.remove(h)
                if len(hour_obj.dyn_ch) > 0:
                    if h in hour_obj.dyn_ch.keys():
                        hour_obj.dyn_ch.pop(h)
        hour_obj.nh.sort()
        
        if len(new_nonhours) > 0 or len(new_ok_hours) > 0:
            conserve_top_up = True

        for h in new_ok_hours:
            if h in hour_obj.nh:
                hour_obj.nh.remove(h)
            elif h in hour_obj.ch:
                hour_obj.ch.remove(h)
                hour_obj.dyn_ch.pop(h)    
        print(hour_obj.nh)
        return HourObject(hour_obj.nh, hour_obj.ch, hour_obj.dyn_ch), conserve_top_up

    @staticmethod
    def _get_average_price(prices_today: list, prices_tomorrow:list) -> float:
        ret = prices_today[14::]
        ret[len(ret):] = prices_tomorrow[0:14]
        return stat.mean(ret)


class HourSelectionHelpers:
    @staticmethod
    def _create_dict(input: list):
        ret = {}
        for idx, val in enumerate(input):
            ret[idx] = val
        print(len(ret))
        if 23 <= len(ret) <= 24:
            return ret
        elif len(ret) == 25:
            _LOGGER.debug(f"Looks like we are heading into DST. re-parsing hours")
            input.pop(2)
            return HourSelectionHelpers._create_dict(input)
        else:
            _LOGGER.exception(f"Could not create dictionary from pricelist: {input} with len {len(ret)}.")
            raise ValueError
        

    @staticmethod
    def _try_parse(input:str, parsetype:type):
        try:
            ret = parsetype(input)
            return ret
        except:
            return False
    
    @staticmethod
    def _convert_none_list(lst: any) -> list:
        ret = []
        if lst is None or not isinstance(lst, list):
            return ret
        try:
            for l in lst:
                if l is None:
                    return ret
            return lst
        except:
            return HourSelectionHelpers._make_array_from_empty(lst)

    @staticmethod
    def _make_array_from_empty(input) -> list:
        array = input.split(",")
        list = [p for p in array if len(p) > 0]
        ret = []
        if len(list) > 24:
            try:
                for l in list:
                    parsed_item = HourSelectionHelpers._try_parse(l, float)
                    if not parsed_item:
                        parsed_item = HourSelectionHelpers._try_parse(l, int)
                    assert type(parsed_item) is float or type(parsed_item) is int
                    ret.append(parsed_item)
                return ret
            except:
                return []
        return []


class HourSelectionCalculations:
    @staticmethod
    def normalize_prices(prices:list) -> list:
        min_price = min(prices)
        c = 0
        if min_price <= 0:
            c = abs(min_price) + 0.01
        ret = []
        for p in prices:
            pp = p+c
            divider = min_price if min_price > 0 else c
            ret.append(pp/divider)
        return ret

    @staticmethod
    def rank_prices(hourdict: dict, normalized_hourdict: dict) -> dict:
        ret = {}
        _maxval = max(hourdict.values())
        _max_normalized = max(normalized_hourdict.values())
        peaqstdev = _maxval/abs(_max_normalized/stat.stdev(normalized_hourdict.values()))
        
        if peaqstdev < min(hourdict.values()):
            peaqstdev = peaqstdev + min(hourdict.values())
        for key in hourdict:
            if hourdict[key] > peaqstdev:
                _permax = round(hourdict[key] / _maxval, 2)
                ret[key] = {"val": hourdict[key], "permax": _permax}
        return HourSelectionCalculations._discard_excessive_hours(ret)

    @staticmethod
    def _discard_excessive_hours(hours: dict):
        """There should always be at least four regular hours before absolute_top_price kicks in."""
        while len(hours) >= 20:
            to_pop = dict(sorted(hours.items(), key=lambda item: item[1]['val']))    
            hours.pop(list(to_pop.keys())[0])
        return hours


from abc import abstractmethod
from dataclasses import dataclass, field
from ...models.chargertype.servicecalls_dto import ServiceCallsDTO
from ...models.chargertype.servicecalls_options import ServiceCallsOptions
from ...models.chargerstates import CHARGERSTATES
from ...models.chargertype.charger_options import ChargerOptions
from ...models.chargertype.charger_entities_model import ChargerEntitiesModel
from .servicecalls import ServiceCalls

CHARGERSTATES_BASE = {
    CHARGERSTATES.Idle: [],
    CHARGERSTATES.Connected: [],
    CHARGERSTATES.Charging: [],
    CHARGERSTATES.Done: []
}

@dataclass
class ChargerBase:
    domainname:str = ""
    native_chargerstates:list = field(default_factory=lambda: [])
    servicecalls:ServiceCalls = None
    chargerstates = CHARGERSTATES_BASE
    entities = ChargerEntitiesModel
    options = ChargerOptions

    def _set_servicecalls(
            self,
            domain: str,
            model: ServiceCallsDTO,
            options: ServiceCallsOptions,
    ) -> None:
        self.servicecalls = ServiceCalls(domain, model, options)

    @abstractmethod
    def validatecharger(self) -> bool:
        pass

    @abstractmethod
    def get_entities(self):
        pass

    @abstractmethod
    def set_sensors(self):
        pass



from enum import Enum

class Dividents(Enum):
    AND = 1
    OR = 2
    UNSET = 3

class SumTypes(Enum):
    Max = 1
    Avg = 2
    Min = 3

class TimePeriods(Enum):
    Hourly = 1
    Daily = 2
    Weekly = 3
    BiWeekly = 4
    Monthly = 5
    Yearly = 6
    UnSet = 7

class CalendarPeriods(Enum):
    Minute = 1
    Hour = 2
    Weekday = 3
    Month = 4
    Quarter = 5

class PriceType(Enum):
    Static = 0
    Tiered = 1






from .threshold import Threshold
from .threshold_lite import ThresholdLite

class ThresholdFactory:
    @staticmethod
    def create(hub):
        if hub.options.peaqev_lite:
            return ThresholdLite(hub)
        return Threshold(hub)